import enc28j60
import time

# notation: _P = position of a field
#           _V = value of a field

# ******* ETH *******
ETH_HEADER_LEN              = 14
# values of certain bytes:
ETHTYPE_ARP_H_V             = 0x08
ETHTYPE_ARP_L_V             = 0x06
ETHTYPE_IP_V                = 0x0800
ETHTYPE_IP_H_V              = 0x08
ETHTYPE_IP_L_V              = 0x00
# byte positions in the ethernet frame:
#
# Ethernet type field (2bytes):
ETH_TYPE_H_P                = 12
ETH_TYPE_L_P                = 13
#
ETH_DST_MAC                 = 0
ETH_SRC_MAC                 = 6


# ******* ARP *******
ETH_ARP_OPCODE_REPLY_H_V    = 0x0
ETH_ARP_OPCODE_REPLY_L_V    = 0x02
#
ETHTYPE_ARP_L_V             = 0x06
# arp.dst.ip
ETH_ARP_DST_IP_P            = 0x26
# arp.opcode
ETH_ARP_OPCODE_H_P          = 0x14
ETH_ARP_OPCODE_L_P          = 0x15
# arp.src.mac
ETH_ARP_SRC_MAC_P           = 0x16
ETH_ARP_SRC_IP_P            = 0x1c
ETH_ARP_DST_MAC_P           = 0x20
ETH_ARP_DST_IP_P            = 0x26

ARP_OPCODE_REQUEST_H_V      = 0x00
ARP_OPCODE_REQUEST_L_V      = 0x01
ARP_OPCODE_REPLY_H_V        = 0x00
ARP_OPCODE_REPLY_L_V        = 0x02

ARP_HARDWARE_TYPE_H_V       = 0x00
ARP_HARDWARE_TYPE_L_V       = 0x01
ARP_PROTOCOL_H_V            = 0x08
ARP_PROTOCOL_L_V            = 0x00
ARP_HARDWARE_SIZE_V         = 0x06
ARP_PROTOCOL_SIZE_V         = 0x04

ARP_HARDWARE_TYPE_H_P       = 0x0E
ARP_HARDWARE_TYPE_L_P       = 0x0F
ARP_PROTOCOL_H_P            = 0x10
ARP_PROTOCOL_L_P            = 0x11
ARP_HARDWARE_SIZE_P         = 0x12
ARP_PROTOCOL_SIZE_P         = 0x13
ARP_OPCODE_H_P              = 0x14
ARP_OPCODE_L_P              = 0x15
ARP_SRC_MAC_P               = 0x16
ARP_SRC_IP_P                = 0x1C
ARP_DST_MAC_P               = 0x20
ARP_DST_IP_P                = 0x26

# ******* IP *******
IP_HEADER_LEN               = 20

IP_PROTO_ICMP_V             = 0x01
IP_PROTO_TCP_V              = 0x06
IP_PROTO_UDP_V              = 0x11
IP_V4_V                     = 0x40
IP_HEADER_LENGTH_V          = 0x05

IP_P                        = 0x0E
IP_HEADER_VER_LEN_P         = 0x0E
IP_TOS_P                    = 0x0F
IP_TOTLEN_H_P               = 0x10
IP_TOTLEN_L_P               = 0x11
IP_ID_H_P                   = 0x12
IP_ID_L_P                   = 0x13
IP_FLAGS_P                  = 0x14     
IP_FLAGS_H_P                = 0x14
IP_FLAGS_L_P                = 0x15
IP_TTL_P                    = 0x16
IP_PROTO_P                  = 0x17
IP_CHECKSUM_P               = 0x18 
IP_CHECKSUM_H_P             = 0x18
IP_CHECKSUM_L_P             = 0x19
IP_SRC_IP_P                 = 0x1A
IP_DST_IP_P                 = 0x1E

IP_SRC_P                    = 0x1a
IP_DST_P                    = 0x1e
IP_HEADER_LEN_VER_P         = 0xe
# ******* ICMP *******
ICMP_TYPE_ECHOREPLY_V       = 0
ICMP_TYPE_ECHOREQUEST_V     = 8
#
ICMP_TYPE_P                 = 0x22
ICMP_CHECKSUM_P             = 0x24

# ******* UDP *******
UDP_HEADER_LEN              = 8
#
UDP_SRC_PORT_H_P            = 0x22
UDP_SRC_PORT_L_P            = 0x23
UDP_DST_PORT_H_P            = 0x24
UDP_DST_PORT_L_P            = 0x25
#
UDP_LEN_H_P                 = 0x26
UDP_LEN_L_P                 = 0x27
UDP_CHECKSUM_H_P            = 0x28
UDP_CHECKSUM_L_P            = 0x29
UDP_DATA_P                  = 0x2a

# ******* TCP *******
#  plain len without the options:
TCP_HEADER_LEN_PLAIN        = 20
        
TCP_FLAG_FIN_V              = 0x01
TCP_FLAGS_FIN_V             = 0x01
TCP_FLAGS_SYN_V             = 0x02
TCP_FLAG_SYN_V              = 0x02
TCP_FLAG_RST_V              = 0x04
TCP_FLAG_PUSH_V             = 0x08
TCP_FLAGS_ACK_V             = 0x10
TCP_FLAG_ACK_V              = 0x10
TCP_FLAG_URG_V              = 0x20
TCP_FLAG_ECE_V              = 0x40
TCP_FLAG_CWR_V              = 0x80
TCP_FLAGS_SYNACK_V          = 0x12

TCP_SRC_PORT_H_P            = 0x22
TCP_SRC_PORT_L_P            = 0x23
TCP_DST_PORT_H_P            = 0x24
TCP_DST_PORT_L_P            = 0x25
TCP_SEQ_P                   = 0x26   # the tcp seq number is 4 bytes 0x26-0x29
TCP_SEQ_H_P                 = 0x26
TCP_SEQACK_P                = 0x2A   # 4 bytes
TCP_SEQACK_H_P              = 0x2A
TCP_HEADER_LEN_P            = 0x2E
TCP_FLAGS_P                 = 0x2F
TCP_FLAG_P                  = 0x2F
TCP_WINDOWSIZE_H_P          = 0x30   # 2 bytes
TCP_WINDOWSIZE_L_P          = 0x31
TCP_CHECKSUM_H_P            = 0x32
TCP_CHECKSUM_L_P            = 0x33
TCP_URGENT_PTR_H_P          = 0x34   # 2 bytes
TCP_URGENT_PTR_L_P          = 0x35
TCP_OPTIONS_P               = 0x36
TCP_DATA_P                  = 0x36

class NET:
    '''
     The Ip checksum is calculated over the ip header only starting
     with the header length field and a total length of 20 bytes
     unitl ip.dst
     You must set the IP checksum field to zero before you start
     the calculation.
     len for ip is 20.
     
     For UDP/TCP we do not make up the required pseudo header. Instead we 
     use the ip.src and ip.dst fields of the real packet:
     The udp checksum calculation starts with the ip.src field
     Ip.src=4bytes,Ip.dst=4 bytes,Udp header=8bytes + data length=16+len
     In other words the len here is 8 + length over which you actually
     want to calculate the checksum.
     You must set the checksum field to zero before you start
     the calculation.
     len for udp is: 8 + 8 + data length
     len for tcp is: 4+4 + 20 + option len + data length
    
     For more information on how this algorithm works see:
     http://www.netfor2.com/checksum.html
     http://www.msc.uky.edu/ken/cs471/notes/chap3.htm
     The RFC has also a C code example: http://www.faqs.org/rfcs/rfc1071.html
    '''
    def __init__(self,mmacaddr, mipaddr, webport):
        self.wwwport = webport
        self.macaddr = mmacaddr
        self.ipaddr = mipaddr
        self.info_hdr_len = 0
        self.info_data_len = 0
        self.seqnum = 0xa               # my initial tcp sequence number
        self.ip_identifier = 1
        self.enc = enc28j60.ENC28J60()

    def Checksum(self, buf, len, type):
        '''
         type 0=ip 
              1=udp
              2=tcp
        '''
        sum = 0
        if type == 1:
            sum += IP_PROTO_UDP_V   # protocol udp
            # the length here is the length of udp (data+header len)
            #  =length given to this function - (IP.scr+IP.dst length)
            sum += (len - 8)        # = real tcp len
        elif type == 2:
            sum += IP_PROTO_TCP_V
            # the length here is the length of tcp (data+header len)
            # =length given to this function - (IP.scr+IP.dst length)
            sum += (len - 8)        # = real tcp len
            # build the sum of 16bit words
        i = 0
        while len > 1:
            sum += 0xFFFF & ((buf[i] << 8) | buf[i+1])
            i += 2
            len -= 2
        # if there is a byte left then add it (padded with zero)
        if  len > 0:
            sum += buf[i] << 8
        # now calculate the sum over the bytes in the sum
        # until the result is only 16bit long
        while sum >> 16:
            sum = (sum & 0xFFFF) + (sum >> 16)
        # build 1's complement:
        return ((sum ^ 0xFFFF) & 0xFFFF)

    def Init_ip_arp_udp_tcp(self):
        '''
         you must call this function once before you use any of the other functions:
        ''' 
        self.enc.Enc28j60Init(self.macaddr)

    def Eth_type_is_arp_and_my_ip(self, buf, len):
        if len < 41:
            return 0
        if (buf[ETH_TYPE_H_P] != ETHTYPE_ARP_H_V) or (buf[ETH_TYPE_L_P] != ETHTYPE_ARP_L_V):
            return 0
        i = 0
        while i < 4:
            if buf[ETH_ARP_DST_IP_P+i] != self.ipaddr[i]:
                return 0
            i += 1
        return 1

    def Eth_type_is_ip_and_my_ip(self, buf, len):
        # eth+ip+udp header is 42
        if len < 42:
            return 0
        if (buf[ETH_TYPE_H_P] != ETHTYPE_IP_H_V) or (buf[ETH_TYPE_L_P]!=ETHTYPE_IP_L_V):
            return 0
        if buf[IP_HEADER_LEN_VER_P] != 0x45:
            # must be IP V4 and 20 byte header
            return 0
        i = 0
        while i < 4:
            if buf[IP_DST_P+i] != self.ipaddr[i]:
                return 0
            i += 1
        return 1

    def Make_eth(self, buf):
        '''
         make a return eth header from a received eth packet
        '''
        # copy the destination mac from the source and fill my mac into src
        i = 0
        while i < 6:
            buf[ETH_DST_MAC + i] = buf[ETH_SRC_MAC + i]
            buf[ETH_SRC_MAC + i] = self.macaddr[i]
            i += 1

    def Make_eth_ip_new(self, buf, dst_mac):
        '''
         make a new eth header for IP packet
        '''
        # copy the destination mac from the source and fill my mac into src
        i = 0
        while i < 6:
            buf[ETH_DST_MAC + i] = dst_mac[i]
            buf[ETH_SRC_MAC + i] = self.macaddr[i]
            i += 1
        buf[ETH_TYPE_H_P] = ETHTYPE_IP_H_V
        buf[ETH_TYPE_L_P] = ETHTYPE_IP_L_V

    def Fill_ip_hdr_checksum(self, buf):
        # clear the 2 byte checksum
        buf[IP_CHECKSUM_P] = 0
        buf[IP_CHECKSUM_P+1] = 0
        buf[IP_FLAGS_P] = 0x40    # don't fragment
        buf[IP_FLAGS_P+1] = 0     # fragement offset
        buf[IP_TTL_P] = 64        # ttl
        # calculate the checksum:
        ck = self.Checksum(buf[IP_P:], IP_HEADER_LEN, 0)
        buf[IP_CHECKSUM_P] = (ck >> 8) & 0xFF
        buf[IP_CHECKSUM_P+1] = ck & 0xff

    def Make_ip_tcp_new(self, buf, len, dst_ip):
        '''
         make a new ip header for tcp packet
         make a return ip header from a received ip packet
        '''
        # set ipv4 and header length
        buf[IP_P] = (IP_V4_V | IP_HEADER_LENGTH_V)
        # set TOS to default 0x00
        buf[IP_TOS_P] = 0
        # set total length
        buf[IP_TOTLEN_H_P] = (len >> 8) & 0xff
        buf[IP_TOTLEN_L_P] = len & 0xff

        # set packet identification
        buf[IP_ID_H_P] = (ip_identifier >> 8) & 0xff
        buf[IP_ID_L_P] = ip_identifier & 0xff
        ip_identifier =+ 1
    
        # set fragment flags
        buf[IP_FLAGS_H_P] = 0x00
        buf[IP_FLAGS_L_P] = 0x00

        # set Time To Live
        buf[IP_TTL_P] = 128

        # set ip packettype to tcp/udp/icmp...
        buf[IP_PROTO_P] = IP_PROTO_TCP_V

        # set source and destination ip address
        i = 0
        while i < 4:
            buf[IP_DST_P + i] = dst_ip[i]
            buf[IP_SRC_P + i] = self.ipaddr[i]
            i += 1
        self.Fill_ip_hdr_checksum(buf)

    def Make_ip(self, buf):
        '''
         make a return ip header from a received ip packet
        '''
        i = 0
        while i < 4:
            buf[IP_DST_P + i] = buf[IP_SRC_P + i]
            buf[IP_SRC_P + i] = self.ipaddr[i]
            i += 1
        self.Fill_ip_hdr_checksum(buf)

    def Make_tcphead(self, buf, rel_ack_num, mss, cp_seq):
        '''
         make a return tcp header from a received tcp packet
         rel_ack_num is how much we must step the seq number received from the
         other side. We do not send more than 255 bytes of text (=data) in the tcp packet.
         If mss=1 then mss is included in the options list

         After calling this function you can fill in the first data byte at TCP_OPTIONS_P+4
         If cp_seq=0 then an initial sequence number is used (should be use in synack)
         otherwise it is copied from the packet we received
        '''
        i = 0
        while i < 2:
            buf[TCP_DST_PORT_H_P + i] = buf[TCP_SRC_PORT_H_P + i]
            i += 1
        # set source port  (http):
        buf[TCP_SRC_PORT_H_P] = self.wwwport>>8     #  source port
        buf[TCP_SRC_PORT_L_P] = self.wwwport & 0xff
        i = 4
        # sequence numbers:
        # add the rel ack num to SEQACK
        while i > 0:
            rel_ack_num = buf[TCP_SEQ_H_P + i - 1] + rel_ack_num
            tseq = buf[TCP_SEQACK_H_P + i - 1]
            buf[TCP_SEQACK_H_P + i - 1] = 0xff & rel_ack_num
            if (cp_seq):
                # copy the acknum sent to us into the sequence number
                buf[TCP_SEQ_H_P + i - 1] = tseq
            else:
                buf[TCP_SEQ_H_P + i - 1] = 0  # some preset vallue
            rel_ack_num = rel_ack_num >> 8
            i -= 1
        if cp_seq == 0:
            # put inital seq number
            buf[TCP_SEQ_H_P + 0] = 0
            buf[TCP_SEQ_H_P + 1] = 0
            # we step only the second byte, this allows us to send packts 
            # with 255 bytes or 512 (if we step the initial seqnum by 2)
            buf[TCP_SEQ_H_P + 2] = self.seqnum
            buf[TCP_SEQ_H_P + 3] = 0
            # step the inititial seq num by something we will not use
            # during this tcp session:
            self.seqnum += 2
        # zero the checksum
        buf[TCP_CHECKSUM_H_P] = 0
        buf[TCP_CHECKSUM_L_P] = 0

        # The tcp header length is only a 4 bit field (the upper 4 bits).
        # It is calculated in units of 4 bytes. 
        # E.g 24 bytes: 24/4=6 => 0x60=header len field
        #buf[TCP_HEADER_LEN_P]=(((TCP_HEADER_LEN_PLAIN+4)/4)) <<4; // 0x60
        if mss > 0:
            # the only option we set is MSS to 1408:
            # 1408 in hex is 0x580
            buf[TCP_OPTIONS_P] = 2
            buf[TCP_OPTIONS_P + 1] = 4
            buf[TCP_OPTIONS_P + 2] = 0x05 
            buf[TCP_OPTIONS_P + 3] = 0x80
            # 24 bytes:
            buf[TCP_HEADER_LEN_P] = 0x60
        else:
            # no options:
            # 20 bytes:
            buf[TCP_HEADER_LEN_P] = 0x50

    def Make_arp_answer_from_request(self, buf):
        i = 0
        self.Make_eth(buf)
        buf[ETH_ARP_OPCODE_H_P] = ETH_ARP_OPCODE_REPLY_H_V
        buf[ETH_ARP_OPCODE_L_P] = ETH_ARP_OPCODE_REPLY_L_V
        # fill the mac addresses:
        while i < 6:
            buf[ETH_ARP_DST_MAC_P + i] = buf[ETH_ARP_SRC_MAC_P + i]
            buf[ETH_ARP_SRC_MAC_P + i] = self.macaddr[i]
            i += 1
        i = 0
        while i < 4:
            buf[ETH_ARP_DST_IP_P + i] = buf[ETH_ARP_SRC_IP_P + i]
            buf[ETH_ARP_SRC_IP_P + i] = self.ipaddr[i]
            i += 1
        # eth+arp is 42 bytes:
        self.enc.Enc28j60PacketSend(42,buf) 

    def Make_echo_reply_from_request(self, buf, len):
        self.Make_eth(buf)
        self.Make_ip(buf)
        buf[ICMP_TYPE_P] = ICMP_TYPE_ECHOREPLY_V
        # we changed only the icmp.type field from request(=8) to reply(=0).
        # we can therefore easily correct the checksum:
        if buf[ICMP_CHECKSUM_P] > (0xff - 0x08):
            buf[ICMP_CHECKSUM_P + 1] += 1
        csum = buf[ICMP_CHECKSUM_P] + 0x08
        buf[ICMP_CHECKSUM_P] = csum & 0xFF
        self.enc.Enc28j60PacketSend(len, buf)

    def Make_udp_reply_from_request(self, buf, data, datalen, port):
        i = 0
        self.Make_eth(buf)
        if datalen > 220:
            datalen = 220
        # total length field in the IP header must be set:
        buf[IP_TOTLEN_H_P] = 0
        buf[IP_TOTLEN_L_P] = IP_HEADER_LEN + UDP_HEADER_LEN + datalen
        self.Make_ip(buf)
        buf[UDP_DST_PORT_H_P] = port >> 8
        buf[UDP_DST_PORT_L_P] = port & 0xff
        # source port does not matter and is what the sender used.
        # calculte the udp length:
        buf[UDP_LEN_H_P] = 0
        buf[UDP_LEN_L_P] = UDP_HEADER_LEN + datalen
        # zero the checksum
        buf[UDP_CHECKSUM_H_P] = 0
        buf[UDP_CHECKSUM_L_P] = 0
        # copy the data:
        while i < datalen:
            buf[UDP_DATA_P + i] = data[i]
            i += 1
        ck = self.Checksum(buf[IP_SRC_P:], 16 + datalen, 1)
        buf[UDP_CHECKSUM_H_P] = ck >> 8
        buf[UDP_CHECKSUM_L_P] = ck & 0xff
        self.enc,Enc28j60PacketSend(UDP_HEADER_LEN + IP_HEADER_LEN + ETH_HEADER_LEN + datalen, buf)

    def Make_tcp_synack_from_syn(self, buf):
        self.Make_eth(buf)
        # total length field in the IP header must be set:
        # 20 bytes IP + 24 bytes (20tcp+4tcp options)
        buf[IP_TOTLEN_H_P] = 0
        buf[IP_TOTLEN_L_P] = IP_HEADER_LEN + TCP_HEADER_LEN_PLAIN + 4
        self.Make_ip(buf)
        buf[TCP_FLAG_P] = TCP_FLAGS_SYNACK_V
        self.Make_tcphead(buf, 1, 1, 0)
        # calculate the checksum, len=8 (start from ip.src) + TCP_HEADER_LEN_PLAIN + 4 (one option: mss)
        ck = self.Checksum(buf[IP_SRC_P:], 8 + TCP_HEADER_LEN_PLAIN + 4, 2)
        buf[TCP_CHECKSUM_H_P] = ck >> 8
        buf[TCP_CHECKSUM_L_P] = ck & 0xff
        # add 4 for option mss:
        self.enc.Enc28j60PacketSend(IP_HEADER_LEN + TCP_HEADER_LEN_PLAIN + 4 + ETH_HEADER_LEN, buf)

    def Get_tcp_data_pointer(self):
        '''
         get a pointer to the start of tcp data in buf
         Returns 0 if there is no data
         You must call init_len_info once before calling this function
        '''
        if self.info_data_len > 0:
            return (TCP_SRC_PORT_H_P + self.info_hdr_len)
        else:
            return 0

    def Init_len_info (self, buf):
        '''
         do some basic length calculations and store the result in static varibales
        '''
        self.info_data_len = (buf[IP_TOTLEN_H_P] << 8) | (buf[IP_TOTLEN_L_P] & 0xff)
        self.info_data_len -= IP_HEADER_LEN
        self.info_hdr_len = (buf[TCP_HEADER_LEN_P] >> 4) * 4  # generate len in bytes;
        self.info_data_len -= self.info_hdr_len
        if self.info_data_len <= 0:
            self.info_data_len = 0

    def Fill_tcp_data_p(self, buf, pos, progmem_s):
        '''
         fill in tcp data at position pos. pos=0 means start of
         tcp data. Returns the position at which the string after
         this string could be filled.
        '''
        # fill in tcp data at position pos
        # with no options the data starts after the checksum + 2 more bytes (urgent ptr)
        for c in progmem_s:
            buf[TCP_CHECKSUM_L_P + 3 + pos] = c
            pos += 1
        return pos

    def Fill_tcp_data(self, buf, pos, s):
        '''
         fill in tcp data at position pos. pos=0 means start of
         tcp data. Returns the position at which the string after
         this string could be filled.
        '''
        # fill in tcp data at position pos
        # with no options the data starts after the checksum + 2 more bytes (urgent ptr)
        for c in s: 
            buf[TCP_CHECKSUM_L_P + 3 + pos] = c
            pos += 1
        return pos

    def Make_tcp_ack_from_any(self, buf):
        '''
         Make just an ack packet with no tcp data inside
         This will modify the eth/ip/tcp header
        ''' 
        self.Make_eth(buf)
        # fill the header:
        buf[TCP_FLAG_P] = TCP_FLAG_ACK_V
        if self.info_data_len == 0:
            # if there is no data then we must still acknoledge one packet
            self.Make_tcphead(buf, 1, 0, 1)                 # no options
        else:
            self.Make_tcphead(buf, self.info_data_len, 0, 1)     # no options
        # total length field in the IP header must be set:
        # 20 bytes IP + 20 bytes tcp (when no options) 
        j = IP_HEADER_LEN + TCP_HEADER_LEN_PLAIN
        buf[IP_TOTLEN_H_P] = j >> 8
        buf[IP_TOTLEN_L_P] = j & 0xff
        self.Make_ip(buf)
        # calculate the checksum, len=8 (start from ip.src) + TCP_HEADER_LEN_PLAIN + data len
        j = self.Checksum(buf[IP_SRC_P:], 8 + TCP_HEADER_LEN_PLAIN, 2)
        buf[TCP_CHECKSUM_H_P] = j >> 8
        buf[TCP_CHECKSUM_L_P] = j & 0xff
        self.enc.Enc28j60PacketSend(IP_HEADER_LEN + TCP_HEADER_LEN_PLAIN + ETH_HEADER_LEN, buf)

    def Make_tcp_ack_with_data(self, buf, dlen):
        '''
         you must have called init_len_info at some time before calling this function
         dlen is the amount of tcp data (http data) we send in this packet
         You can use this function only immediately after make_tcp_ack_from_any
         This is because this function will NOT modify the eth/ip/tcp header except for
         length and checksum
        '''
        # fill the header:
        # This code requires that we send only one data packet
        # because we keep no state information. We must therefore set
        # the fin here:
        buf[TCP_FLAG_P] = TCP_FLAG_ACK_V | TCP_FLAG_PUSH_V | TCP_FLAG_FIN_V
        # total length field in the IP header must be set:
        # 20 bytes IP + 20 bytes tcp (when no options) + len of data
        j = IP_HEADER_LEN + TCP_HEADER_LEN_PLAIN + dlen
        buf[IP_TOTLEN_H_P] = j >> 8
        buf[IP_TOTLEN_L_P] = j & 0xff
        self.Fill_ip_hdr_checksum(buf)
        # zero the checksum
        buf[TCP_CHECKSUM_H_P] = 0
        buf[TCP_CHECKSUM_L_P] = 0
        # calculate the checksum, len=8 (start from ip.src) + TCP_HEADER_LEN_PLAIN + data len
        j = self.Checksum(buf[IP_SRC_P:], 8 + TCP_HEADER_LEN_PLAIN + dlen, 2)
        buf[TCP_CHECKSUM_H_P] = j >> 8
        buf[TCP_CHECKSUM_L_P] = j & 0xff
        self.enc.Enc28j60PacketSend(IP_HEADER_LEN + TCP_HEADER_LEN_PLAIN + dlen + ETH_HEADER_LEN, buf)

    def Tcp_get_dlength (self, buf):
        dlength = (buf[IP_TOTLEN_H_P] << 8) | buf[IP_TOTLEN_L_P]
        dlength -= IP_HEADER_LEN
        hlength = (buf[TCP_HEADER_LEN_P] >> 4) * 4   # generate len in bytes;
        dlength -= hlength
        if dlength <= 0:
            dlength=0
        return dlength

    def isEchoReply(self, buf):
        return (buf[IP_PROTO_P] == IP_PROTO_ICMP_V and buf[ICMP_TYPE_P] == ICMP_TYPE_ECHOREQUEST_V)
        
    def isTcpPort(self, buf):
        return (buf[IP_PROTO_P] == IP_PROTO_TCP_V and buf[TCP_DST_PORT_H_P] == ((self.wwwport>>8) & 0xFF) and buf[TCP_DST_PORT_L_P] == (self.wwwport & 0xFF))

    def isSYNflag(self, buf):
        return (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V)
            
    def isACKflag(self, buf):
        return (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V)

    def isFINflag(self, buf):
        return (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V)

